import multiprocessing as mp
import numpy as np
from collections import OrderedDict
import copy


class development_tree(object):
    def __init__(self, w, cell_types, parent_types, frac_tf_per_branch=0.01):
        self.w = w
        self.cell_types = cell_types
        self.parent_types = OrderedDict(zip(cell_types, parent_types))
        self.n = w.shape[0]
        self.frac_tf_per_branch = frac_tf_per_branch
        self.n_tf_per_branch = int(frac_tf_per_branch * self.n)

        self.get_stem_cell_type()
        self.get_tfs()
        self.get_paths()
        self.get_equilibrium_ts()
        self.get_s0()

    def get_stem_cell_type(self):
        self.stem_cell_type = None
        for cell_type, parent_type in self.parent_types.iteritems():
            if parent_type is None:
                self.stem_cell_type = cell_type
                break
        if self.stem_cell_type is None:
            print 'development_tree: At least one parent cell type must be None.'
            exit

    def get_tfs(self):

        self.regulators = np.where(np.abs(self.w).sum(0) > 0)[1]
        self.tfs = {}

        def __get_tf(cell_type):
            if cell_type == self.stem_cell_type:
                tf = np.random.choice(self.regulators, self.n_tf_per_branch)
                sign = np.sign(np.array(self.w[:, tf].sum(0))[0])
                self.tfs[cell_type] = dict(zip(tf, sign))
            else:
                old_tf = list(self.tfs[self.parent_types[cell_type]].keys())
                unused_regulators = list(set(self.regulators) - set(old_tf))
                new_tf = list(
                    np.random.choice(
                        unused_regulators,
                        self.n_tf_per_branch, ))
                # old_tf = list(
                #     np.random.choice(
                #         old_tf,
                #         int(len(old_tf) * (1 - self.frac_tf_per_branch)),
                #         replace=False))
                tf = old_tf + new_tf
                sign = np.sign(np.array(self.w[:, tf].sum(0))[0])
                self.tfs[cell_type] = dict(zip(tf, sign))
            for child_type, parent_type in self.parent_types.iteritems():
                if parent_type == cell_type:
                    __get_tf(child_type)

        __get_tf(self.stem_cell_type)

    def get_paths(self):
        def __get_path(cell_type, path):
            path.insert(0, cell_type)
            if cell_type == self.stem_cell_type:
                return path
            for child_type, parent_type in self.parent_types.iteritems():
                if child_type == cell_type:
                    return __get_path(parent_type, path)

        self.paths = {}
        for cell_type in self.cell_types:
            self.paths[cell_type] = __get_path(cell_type, [])

    def get_equilibrium_ts(self, t0=0.0, dt=0.1, tn=1e3, eps=1e-5, n_ic=5):

        t = np.arange(t0, tn, dt)
        nt = t.shape[0]

        self.equilibrium_ts = np.empty(len(self.cell_types))

        def __get_equilibrium_t(tf={}):
            t_inf = np.empty(n_ic)
            s = np.empty((self.n, nt + 1))
            for iic in range(n_ic):
                s[:, 0] = np.random.uniform(-1, 1, size=self.n)
                for gene, sign in tf.iteritems():
                    s[gene, 0] = sign
                for it, ti in enumerate(t):
                    s[:, it + 1] = s[:, it] + (
                        -s[:, it] + np.tanh(self.w * s[:, it])) * dt
                    for gene, sign in tf.iteritems():
                        s[gene, it + 1] = sign
                    if np.linalg.norm(s[:, it + 1] - s[:, it]
                                      ) / np.linalg.norm(s[:, it]) < eps:
                        t_inf[iic] = ti
                        break
                if ti == t[-1]:
                    break
            t_inf = t_inf.mean() if ti != t[-1] else tn
            return t_inf

        self.equilibrium_ts = {}
        for cell_type in self.cell_types:
            self.equilibrium_ts[cell_type] = __get_equilibrium_t(
                tf=self.tfs[cell_type])

    def get_s0(self, t0=0.0, dt=0.1, tn=1e3, eps=1e-5):

        t = np.arange(t0, tn, dt)
        nt = t.shape[0]

        s = np.random.uniform(-1, 1, size=self.n)
        for gene, sign in self.tfs['stem'].iteritems():
            s[gene] = sign
        for it, ti in enumerate(t):
            next_s = s + (-s + np.tanh(self.w * s)) * dt
            for gene, sign in self.tfs['stem'].iteritems():
                next_s[gene] = sign
            if np.linalg.norm(next_s - s) / np.linalg.norm(s) < eps:
                break
            s = next_s.copy()
        self.s0 = s


def simulate_cell((icell, cell, dev_tree, nt, dt, dw_scale, tf_weight,
                   verbose)):

    w = dev_tree.w
    n = w.shape[0]

    s = [dev_tree.s0]
    path = dev_tree.paths[cell]

    for ibranch, branch in enumerate(path):

        tn = dt * (nt[ibranch] + 1)
        t = np.arange(0.0, tn, dt)

        for it, ti in enumerate(t):
            dw = np.random.normal(loc=0, scale=dt, size=n)
            s.append(s[-1] +
                     (-s[-1] + np.tanh(w * s[-1] + dw_scale * dw)) * dt)
            for gene, sign in dev_tree.tfs[branch].iteritems():
                s[-1][gene] = (
                    1.0 - tf_weight * dt) * s[-1][gene] + tf_weight * dt * sign

    if verbose:
        print 'cell:', icell, 'type:', cell, 'age:', len(s)

    return np.array(s)


def simulate_scRNAseq_data(dev_tree,
                           n_cells,
                           time_dist='uniform',
                           dt=0.1,
                           dw_scale=0.1,
                           tf_weight=0.5,
                           frac_dropout=0.0,
                           r_dispersion=False,
                           r_loc=1.0,
                           r_scale=0.1,
                           n_processes=None,
                           verbose=False):

    cells = np.repeat(dev_tree.cell_types, n_cells)

    kwds = (dt, dw_scale, tf_weight, verbose)

    if time_dist == 'uniform':
        dist = np.random.uniform
    elif time_dist == 'poisson':
        dist = np.random.poisson
    nt = [[
        dist(dev_tree.equilibrium_ts[branch] // dt)
        for branch in dev_tree.paths[cell]
    ] for cell in cells]

    args = ((icell, cell, copy.deepcopy(dev_tree), nt[icell]) + kwds
            for icell, cell in enumerate(cells))

    if n_processes is None:
        n_processes = mp.cpu_count()
    pool = mp.Pool(processes=n_processes, maxtasksperchild=16)

    res = pool.map_async(simulate_cell, args)
    res.wait()

    history = res.get()

    pool.close()
    pool.terminate()
    pool.join()

    ages = np.array([len(expr) for expr in history], dtype=float)
    ages -= ages.min()
    ages /= ages.max()

    data = np.array([expr[-1] for expr in history])

    n_genes = data.shape[1]

    # dropout and radius dispersion
    n_dropout = int(frac_dropout * n_genes)
    for irow, row in enumerate(data):
        dropout = np.random.choice(n_genes, size=n_dropout)
        row[dropout] = 0.0
        if r_dispersion:
            r = np.random.normal(loc=r_loc, scale=r_scale)
            r = max(0, r)
            row *= r * r * r
        data[irow] = row

    # cols: cells, rows: genes
    data = data.T

    return data, history, cells, ages



from sklearn.decomposition import PCA
from matplotlib.animation import ArtistAnimation
import matplotlib.pyplot as plt
import numpy as np


def plot_scRNAseq_data(data, cells=None, t=None, color_by='time'):

    data_pca = PCA(n_components=2).fit_transform(data.T)

    fig = plt.figure()
    ax = plt.gca()

    if color_by == 'time':
        c = t
        cmap = plt.get_cmap('Reds')
    elif color_by == 'cell_type':
        cell_types = np.unique(cells)
        c = np.array(
            [np.where(cell_types == cell)[0][0] for cell in cells],
            dtype=float)
        c /= c.max()
        cmap = plt.get_cmap('viridis')

    ax.scatter(*data_pca.T, c=c, cmap=cmap)
    ax.set_xlabel('Component 1')
    ax.set_ylabel('Component 2')

    return fig

def animate_scRNAseq_data(data,
                          cells=None,
                          t=None,
                          n_frames=100,
                          color_by='cell_type'):

    fig = plt.figure(figsize=(6, 6))
    ax = fig.gca()
    ax.set_xlabel('Component 1')
    ax.set_ylabel('Component 2')

    artists = []

    data_final = np.array([expr[-1] for expr in data])
    pca = PCA(n_components=2).fit(data_final)

    if color_by == 'time':
        c = t
        cmap = plt.get_cmap('Reds')
    elif color_by == 'cell_type':
        cell_types = np.unique(cells)
        c = np.array(
            [np.where(cell_types == cell)[0][0] for cell in cells],
            dtype=float)
        c /= c.max()
        cmap = plt.get_cmap('viridis')

    ages = np.array([len(expr) for expr in data])

    for age in np.linspace(0, ages.max(), n_frames, dtype=int):
        data_t = np.array([expr[min(age, len(expr) - 1)] for expr in data])
        data_t_pca = pca.transform(data_t)
        artists.append([ax.scatter(*data_t_pca.T, c=c, cmap=cmap)])

    anim = ArtistAnimation(fig, artists, interval=60, blit=True)

    return anim

def plot_pseudotime_tree(tr, data, cells=None, t=None, color_by='time'):

    if color_by == 'time':
        c = t
        cmap = plt.get_cmap('Reds')
    elif color_by == 'cell_type':
        cell_types = np.unique(cells)
        c = np.array(
            [np.where(cell_types == cell)[0][0] for cell in cells],
            dtype=float)
        c /= c.max()
        cmap = plt.get_cmap('viridis')

    pca = PCA(n_components=2).fit(np.hstack((data, tr.v)).T)
    data_pca = pca.transform(data.T)

    fig = plt.figure(figsize=(8, 8))
    ax = plt.gca()
    ax.scatter(*data_pca.T, c=c, cmap=cmap)

    v_pca = pca.transform(tr.v.T)
    idxs = np.stack(np.where(tr.a), 1)
    for idx1, idx2 in idxs:
        x1, y1 = v_pca[idx1]
        x2, y2 = v_pca[idx2]
        ax.plot([x1, x2], [y1, y2], 'k-', lw=0.5)
    ax.scatter(*v_pca.T, color='k', s=2)
    ax.scatter(*v_pca[[tr.root]].T, color='k', s=200)

    ax.set_xlabel('Component 1')
    ax.set_ylabel('Component 2')

    return fig



import numpy as np
import multiprocessing as mp
from sklearn.linear_model import LassoCV
from scipy.sparse import lil_matrix

def warn(*args, **kwargs):
    pass

import warnings
warnings.warn = warn

def fit_grn_row((i, x, y, eps, max_iter, verbose)):
    model = LassoCV(eps=eps, max_iter=max_iter).fit(x, y)
    if verbose:
	print 'row:', i, 'nnz:', (~np.isclose(model.coef_, 0)).sum(), 'score:', model.score(x, y), 'reg param', model.alpha_
    return model.coef_


def fit_grn(data, t, dt=0.1, dt_eps=1e-3, eps=1e-2, max_iter=500, n_processes=None, verbose=False):

    order = np.argsort(t)
    t_sorted = t[order]
    x = data[:, order]

    pairs = []
    for it, ti in enumerate(t_sorted):
        tmp = np.abs(t_sorted - (ti + dt))
        idx = np.argmin(tmp)
        if tmp[idx] < dt_eps:
            pairs.append([it, idx])
    pairs = np.array(pairs).T

    x1 = x[:, pairs[0]]
    x2 = x[:, pairs[1]]
    dx = x2 - x1
    dt = t_sorted[pairs[1]] - t_sorted[pairs[0]]
    dt /= dt.min()
    y = x1 + dx / dt
    y = np.arctanh(y)
    x1 = x1.T

    if n_processes is None:
        n_processes = mp.cpu_count()
    pool = mp.Pool(processes=n_processes)

    kwds = (eps, max_iter, verbose)
    args = [(iy, x1, yi) + kwds for iy, yi in enumerate(y)]

    res = pool.map_async(fit_grn_row, args)
    res.wait()

    w_rows = res.get()

    pool.close()
    pool.terminate()
    pool.join()

    n = data.shape[0]
    w_fit = lil_matrix((n, n))
    for irow, row in enumerate(w_rows):
        w_fit[irow] = row

    return w_fit.tocsr()

def grn_structure_score(w, w_fit):

    # def csr_ij(w):
    #     ij = []
    #     for row, (ptr1, ptr2) in enumerate(zip(w.indptr[:-1], w.indptr[1:])):
    #         for col in w.indices[ptr1:ptr2]:
    #         ij.append((row,col))
    #     return ij
    # w_ij = set(csr_ij(w))
    # w_fit_ij = set(csr_ij(w_fit))
    # common = list(w_ij.intersection(w_fit_ij))
    # uncommon = [list(w_ij-common), list(w_fit_ij-common)]
    # m = len(common), len(uncommon[0]), len(uncommon[1])
    # score = (n*n - np.sum(m)) + m[0] - m[1] - m[2]
    # score /= float(n*n)
    # return score

    return (w.toarray().astype(bool) == w_fit.toarray().astype(bool)).mean()






import time
import matplotlib.pyplot as plt
from scipy.sparse.csgraph import minimum_spanning_tree, laplacian
import numpy as np
from sklearn.cluster import KMeans
import tree

reload(tree)
from sklearn.decomposition import PCA


# reversed graph embedding
def rge(x, c, max_iter=100, eps=1e-5, sig=0.1, gam=1.0, verbose=False):

    START = time.time()
    if verbose:
        print 'RGE'
        plt.ion()

    objs = []

    for it in range(max_iter):

        phi = (c**2).sum(0)[:, np.newaxis] - 2.0 * c.T.dot(c) + (c**2).sum(0)
        phi[np.diag_indices_from(phi)] = 0.0
        stree = minimum_spanning_tree(phi).todense()
        w = stree + stree.T
        w = w.astype(bool).astype(np.float64)
        obj_w = stree.sum()
        dist_xc = (x**2).sum(0)[:, np.newaxis] - 2.0 * x.T.dot(c) + (c**
                                                                     2).sum(0)
        min_dist = dist_xc.min(1)
        dist_xc -= min_dist[:, np.newaxis]
        phi_xc = np.exp(-dist_xc / sig)
        p = phi_xc / phi_xc.sum(1)[:, np.newaxis]
        obj_p = -sig * (np.log(phi_xc.sum(1)) - min_dist / sig).sum()

        obj = obj_w + gam * obj_p
        objs.append(obj)

        if verbose:
            print 'iter', it, 'obj =', obj
        if it > 0:
            relative_diff = abs(objs[it - 1] - obj) / abs(objs[it - 1])
            if relative_diff < eps:
                if verbose:
                    print 'eps =', relative_diff, 'converge'
                break
        if it >= max_iter:
            if verbose:
                print 'eps =', relative_diff, 'reached max iters'
            break

        q = -w.copy()
        q[np.diag_indices_from(q)] += w.sum(0)
        q *= 2.0
        q[np.diag_indices_from(q)] += gam * p.sum(0)
        b = gam * x.dot(p)
        c = np.linalg.solve(q.T, b.T).T

    if verbose:
        print(w.sum(0) == 1).sum(), 'leaves'
        print 'TOTAL', round(time.time() - START, 2), 's'

    return c, stree


def pseudotime(data,
               cells,
               var_explained=0.99,
               m=None,
               k=None,
               resid_factor=1.0,
               max_iter=1,
               eps=1e-5,
               sig=0.1,
               gam=1.0,
               verbose=False):

    x = data.copy().astype(np.float64)
    if m is None:
        pca_evrcs = PCA().fit(x.T).explained_variance_ratio_.cumsum()
        m = (pca_evrcs < var_explained).sum()
        m = max(2, m)
    pca = PCA(n_components=m)
    x = pca.fit_transform(x.T).T
    ev = pca.explained_variance_ratio_.cumsum()[m - 1]
    if verbose:
        print 'pca, n_components =', m, 'var explained', round(ev, 3)

    shift = x.mean(1)[:, np.newaxis]
    x -= shift
    scale = np.abs(x).max(1)[:, np.newaxis]
    x /= scale

    # k means
    n = max(x.shape)
    start = time.time()
    if k is None:
        k = n

    c = KMeans(n_clusters=k).fit(x.T).cluster_centers_.T

    if verbose:
        print 'k-means, k =', k, round(time.time() - start, 2), 's'

    for it in range(max_iter):

        last_x = x.copy()
        last_c = c.copy()

        c, stree = rge(x, c, eps=eps, sig=sig, gam=gam, verbose=verbose)

        e_idx = np.stack(np.where(stree))
        e = c[:, e_idx]
        x2 = (x * x).sum(0)[:, np.newaxis]
        d = x2 - 2.0 * x.T.dot(e[:, 0]) + (e[:, 0]**2).sum(0)
        d += x2 - 2.0 * x.T.dot(e[:, 1]) + (e[:, 1]**2).sum(0)
        idx = d.argmin(1)

        a = np.diff(e[..., idx], axis=1).squeeze()
        b = x - e[:, 0, idx]

        proj = []
        for i, (ai, bi) in enumerate(zip(a.T, b.T)):
            proj.append(ai.dot(bi) / ai.dot(ai))
            r = bi - proj[-1] * ai
            bi -= resid_factor * r
            x[:, i] = bi + e[:, 0, idx[i]]

        proj = np.array(proj)
        e = e_idx[:, idx].T

        c_err = ((c - last_c)**2).sum(0) / (c**2).sum(0)
        x_err = ((x - last_x)**2).sum(0) / (x**2).sum(0)

        if verbose:
            print it, c_err.mean(), x_err.mean()

        if c_err.mean() < eps and x_err.mean() < eps:
            break

    stree = np.array(stree)

    x, c = scale * x, scale * c
    x, c = shift + x, shift + c

    x = pca.inverse_transform(x.T).T
    c = pca.inverse_transform(c.T).T

    if verbose:
        fig = plt.figure(figsize=(8, 8))
        ax = plt.gca()
        pca = PCA(n_components=2).fit(np.hstack((x, c)).T)
        x_pca = pca.transform(x.T)
        c_pca = pca.transform(c.T)
        ax.scatter(*x_pca.T, color='k')
        idxs = np.stack(np.where(stree), 1)
        for idx1, idx2 in idxs:
            x1, y1 = c_pca[idx1]
            x2, y2 = c_pca[idx2]
            ax.plot([x1, x2], [y1, y2], 'k-', lw=0.5)
        ax.scatter(*c_pca.T, color='k', s=2)
        ax.axis('off')
        plt.show()

    tr = tree.tree(c, stree)

    stem_cell_data = data[:, cells == 'stem']
    tr.set_root(stem_cell_data)

    tr.get_pseudotime()

    t = []
    for ie, ei in enumerate(e):
        v1, v2 = ei
        t1, t2 = tr.t[v1], tr.t[v2]
        t.append(t1 + proj[ie] * (t2 - t1))

    t = np.array(t)
    t -= t.min()
    t /= t.max()

    return t, tr



import numpy as np


class tree(object):
    def __init__(self, v, w):
        self.v = v
        self.w = w
        self.a = (w + w.T).astype(bool).astype(int)
        self.leaves = np.where(self.a.sum(1) == 1)[0]
        self.branch_points = np.where(self.a.sum(1) > 2)[0]
        self.root = None

    def set_root(self, stem_cell_data=None):
        leaf_data = self.v[:, self.leaves]
        dist = (leaf_data**2).sum(0)[:, np.newaxis] - 2.0 * leaf_data.T.dot(
            stem_cell_data) + (stem_cell_data**2).sum(0)
        self.root = self.leaves[np.argmin(dist.sum(1))]

    def get_branch(self, start):
        branch = [start[0], start[1]]
        e = [start[0], start[1]]
        starts = []
        while True:
            ee = np.where(self.a[e[-1]])[0]
            eee = list(set(ee) - {e[-2]})

            if e[-1] in self.branch_points:  # len(ee) > 2
                # print start, 'branch point'
                starts = np.stack((np.repeat(e[-1], len(eee)), eee), 1)
                break

            if e[-1] in self.leaves:  # len(ee) = 1
                # print 'start', 'leaf'
                break

            e = [e[-1], eee[0]]

            branch.append(e[-1])
            # print start, branch

        self.branches.append(branch)

        for start in starts:
            self.get_branch(start)

    def get_pseudotime(self):
        if self.root is None:
            # print 'root is not defined, choosing at random'
            self.root = np.random.choice(self.leaves)
        self.branches = []
        start = [self.root, np.where(self.a[self.root])[0][0]]
        self.get_branch(start)

        self.dt = [
            # np.repeat(1.0, len(branch))
            np.sqrt((np.diff(self.v[:, branch])**2).sum(0))
            for branch in self.branches
        ]

        self.t = np.zeros(self.a.shape[0])
        for branch, dti in zip(self.branches, self.dt):
            for i in range(len(branch) - 1):
                self.t[branch[i + 1]] = self.t[branch[i]] + dti[i]



from setuptools import setup

setup(
    name='RNAscrutiny',
    version='1.2',
    description='Tool for simulating and analyzing single-cell RNA-sequencing data.',
    keywords=
    'RNA single cell simulation inference generative genomics transcriptomics',
    url='http://lbm.niddk.nih.gov/mckennajp/scRutiNy',
    author='Joseph P McKena',
    author_email='joseph.mckenna@nih.gov',
    license='MIT',
    packages=['RNAscrutiny'],
    install_requires=['numpy', 'matplotlib', 'sklearn'],
    zip_safe=False)



import pandas as pd
import numpy as np
from scipy.sparse import lil_matrix

def get_trrust_grn(organism='human'):

    raw_data = pd.read_csv(
        'http://www.grnpedia.org/trrust/data/trrust_rawdata.%s.tsv' %
        (organism, ),
        sep='\t',
        usecols=[0, 1, 2],
        header=None)

    genes = np.unique(raw_data.iloc[:, :2].values.flatten())
    n = genes.shape[0]

    w = lil_matrix((n, n))

    for row, (out_gene, in_gene, sign) in raw_data.iterrows():
        i = np.where(genes == in_gene)[0][0]
        j = np.where(genes == out_gene)[0][0]
        if sign == 'Activation':
            wij = 1.0
            # wij = np.random.uniform(0, 1)
        elif sign == 'Repression':
            wij = -1.0
            # wij = np.random.uniform(-1, 0)
        else:
            wij = np.random.uniform(-1, 1)
        w[i, j] = wij

    return w.tocsr(), genes

def get_grn(name='TRRUST', organism='human'):
    if name == 'TRRUST':
        w, genes = get_trrust_grn(organism=organism)
    return w, genes
